from typing import Optional
from urllib.parse import urljoin


def get_owner_url(owner: str) -> str:
    return "/{}".format(owner)


def get_project_url(unique_name: str) -> str:
    values = unique_name.split(".")
    return "{}/{}".format(get_owner_url(values[0]), values[1])


def get_owner_project_url(owner: str, project_name: str) -> str:
    return "{}/{}".format(get_owner_url(owner), project_name)


def get_fqn_run_url(unique_name: str) -> str:
    values = unique_name.split(".")
    project_url = get_owner_project_url(owner=values[0], project_name=values[1])
    return f"{project_url}/runs/{values[-1]}"


def get_run_url(owner: str, project_name: str, run_uuid: str) -> str:
    project_url = get_owner_project_url(owner=owner, project_name=project_name)
    return f"{project_url}/runs/{run_uuid}"


def get_run_health_url(unique_name: str) -> str:
    run_url = get_fqn_run_url(unique_name=unique_name)
    return f"{run_url}/_heartbeat"


def get_proxy_run_url(
    service: str,
    namespace: str,
    owner: str,
    project: str,
    run_uuid: str,
    subpath: Optional[str] = None,
    port: Optional[int] = None,
) -> str:
    url_path = "{namespace}/{owner}/{project}/runs/{run_uuid}".format(
        namespace=namespace,
        owner=owner,
        project=project,
        run_uuid=run_uuid,
    )
    if port:
        url_path = "{}/{}".format(url_path, port)
    if subpath:
        url_path = "{}/{}".format(url_path, subpath)
    return urljoin(service.rstrip() + "/", url_path.lstrip("/"))


def get_run_reconcile_url(unique_name: str) -> str:
    run_url = get_fqn_run_url(unique_name=unique_name)
    return "{}/_reconcile".format(run_url)


URL_FORMAT = "{protocol}://{domain}{path}"



from typing import Dict, Iterable, List, Optional

from clipped.utils.lists import to_list

from polyaxon._connections import V1Connection, V1ConnectionResource
from polyaxon._docker import docker_types
from polyaxon._flow import V1Init, V1Plugins
from polyaxon._runner.converter import BaseConverter as _BaseConverter
from polyaxon.exceptions import PolyaxonConverterError


class MainConverter(_BaseConverter):
    def _get_main_container(
        self,
        container_id: str,
        main_container: docker_types.V1Container,
        plugins: V1Plugins,
        artifacts_store: Optional[V1Connection],
        init: Optional[List[V1Init]],
        connections: Optional[List[str]],
        connection_by_names: Dict[str, V1Connection],
        secrets: Optional[Iterable[V1ConnectionResource]],
        config_maps: Optional[Iterable[V1ConnectionResource]],
        run_path: Optional[str],
        kv_env_vars: List[List] = None,
        ports: List[int] = None,
    ) -> docker_types.V1Container:
        connections = connections or []
        connection_by_names = connection_by_names or {}
        secrets = secrets or []
        config_maps = config_maps or []

        if artifacts_store and not run_path:
            raise PolyaxonConverterError("Run path is required for main container.")

        if artifacts_store and (
            not plugins.collect_artifacts or plugins.mount_artifacts_store
        ):
            if artifacts_store.name not in connection_by_names:
                connection_by_names[artifacts_store.name] = artifacts_store
            if artifacts_store.name not in connections:
                connections.append(artifacts_store.name)

        requested_connections = [connection_by_names[c] for c in connections]
        requested_config_maps = V1Connection.get_requested_resources(
            resources=config_maps,
            connections=requested_connections,
            resource_key="config_map",
        )
        requested_secrets = V1Connection.get_requested_resources(
            resources=secrets, connections=requested_connections, resource_key="secret"
        )

        # Mounts
        volume_mounts = (
            self._get_mounts(
                use_auth_context=plugins.auth,
                use_artifacts_context=False,  # Main container has a check and handling for this
                use_docker_context=plugins.docker,
                use_shm_context=plugins.shm,
                run_path=run_path,
            )
            if plugins
            else []
        )
        volume_mounts = volume_mounts + self._get_main_volume_mounts(
            plugins=plugins,
            init=init,
            connections=requested_connections,
            secrets=requested_secrets,
            config_maps=requested_config_maps,
            run_path=run_path,
        )

        # Env vars
        env = self._get_main_env_vars(
            plugins=plugins,
            kv_env_vars=kv_env_vars,
            artifacts_store_name=artifacts_store.name if artifacts_store else None,
            connections=requested_connections,
            secrets=requested_secrets,
            config_maps=requested_config_maps,
        )

        # Env from
        resources = to_list(requested_secrets, check_none=True) + to_list(
            requested_config_maps, check_none=True
        )
        env += self._get_env_from_json_resources(resources=resources)

        ports = to_list(ports, check_none=True)

        return self._patch_container(
            container=main_container,
            name=container_id,
            env=env,
            volume_mounts=volume_mounts,
            ports=ports or None,
        )



from typing import Optional

from kubernetes import client, config
from kubernetes.client import Configuration
from kubernetes.client.rest import ApiException

from polyaxon._k8s import constants
from polyaxon._k8s.manager.base import BaseK8sManager
from polyaxon._k8s.monitor import is_pod_running
from polyaxon.exceptions import PolyaxonK8sError
from polyaxon.logger import logger


class K8sManager(BaseK8sManager):
    CLIENT = client

    def __init__(self, namespace="default", in_cluster=False, k8s_config=None):
        super().__init__(namespace=namespace, in_cluster=in_cluster)
        if not k8s_config:
            if in_cluster:
                config.load_incluster_config()
            else:
                config.load_kube_config()
            self.api_client = None
        else:
            self.api_client = client.api_client.ApiClient(configuration=k8s_config)

    @staticmethod
    async def load_config(
        in_cluster: bool = False, k8s_config: Optional[bool] = None
    ) -> Configuration:
        if not k8s_config:
            if in_cluster:
                config.load_incluster_config()
            else:
                await config.load_kube_config()
        return Configuration.get_default_copy()

    def get_version(self, reraise=False):
        try:
            return self.k8s_version_api.get_code().to_dict()
        except ApiException as e:
            logger.error("K8S error: {}".format(e))
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e

    def _list_namespace_resource(self, resource_api, reraise=False, **kwargs):
        try:
            res = resource_api(namespace=self.namespace, **kwargs)
            if isinstance(res, dict):
                items = res["items"]
            else:
                items = res.items
            return [p for p in items]
        except ApiException as e:
            logger.error("K8S error: {}".format(e))
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return []

    def list_nodes(self, reraise=False):
        try:
            res = self.k8s_api.list_node()
            return [p for p in res.items]
        except ApiException as e:
            logger.error("K8S error: {}".format(e))
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return []

    def list_pods(self, include_uninitialized=True, reraise=False, **kwargs):
        return self._list_namespace_resource(
            resource_api=self.k8s_api.list_namespaced_pod,
            reraise=reraise,
            **kwargs,
        )

    def list_jobs(self, include_uninitialized=True, reraise=False, **kwargs):
        return self._list_namespace_resource(
            resource_api=self.k8s_batch_api.list_namespaced_job,
            reraise=reraise,
            **kwargs,
        )

    def list_custom_objects(self, group, version, plural, reraise=False, **kwargs):
        return self._list_namespace_resource(
            resource_api=self.k8s_custom_object_api.list_namespaced_custom_object,
            reraise=reraise,
            group=group,
            version=version,
            plural=plural,
            **kwargs,
        )

    def list_services(self, reraise=False, **kwargs):
        return self._list_namespace_resource(
            resource_api=self.k8s_api.list_namespaced_service,
            reraise=reraise,
            **kwargs,
        )

    def list_deployments(self, reraise=False, **kwargs):
        return self._list_namespace_resource(
            resource_api=self.k8s_apps_api.list_namespaced_deployment,
            reraise=reraise,
            **kwargs,
        )

    def list_ingresses(self, reraise=False, **kwargs):
        return self._list_namespace_resource(
            resource_api=self.networking_v1_beta1_api.list_namespaced_ingress,
            reraise=reraise,
            **kwargs,
        )

    def update_node_labels(self, node, labels, reraise=False):
        body = {"metadata": {"labels": labels}, "namespace": self.namespace}
        try:
            return self.k8s_api.patch_node(name=node, body=body)
        except ApiException as e:
            logger.error("K8S error: {}".format(e))
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e

    def create_config_map(self, name, body):
        resp = self.k8s_api.create_namespaced_config_map(
            namespace=self.namespace, body=body
        )
        logger.debug("Config map `{}` was created".format(name))
        return resp

    def update_config_map(self, name, body):
        resp = self.k8s_api.patch_namespaced_config_map(
            name=name, namespace=self.namespace, body=body
        )
        logger.debug("Config map `{}` was patched".format(name))
        return resp

    def create_or_update_config_map(self, name, body, reraise=False):
        try:
            return self.create_config_map(name=name, body=body)
        except ApiException:
            try:
                return self.update_config_map(name=name, body=body)
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_secret(self, name, body):
        resp = self.k8s_api.create_namespaced_secret(
            namespace=self.namespace, body=body
        )
        logger.debug("Secret `{}` was created".format(name))
        return resp

    def update_secret(self, name, body):
        resp = self.k8s_api.patch_namespaced_secret(
            name=name, namespace=self.namespace, body=body
        )
        logger.debug("Secret `{}` was patched".format(name))
        return resp

    def create_or_update_secret(self, name, body, reraise=False):
        try:
            return self.create_secret(name=name, body=body), True
        except ApiException:
            try:
                return self.update_secret(name=name, body=body), False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_service(self, name, body):
        resp = self.k8s_api.create_namespaced_service(
            namespace=self.namespace, body=body
        )
        logger.debug("Service `{}` was created".format(name))
        return resp

    def update_service(self, name, body):
        resp = self.k8s_api.patch_namespaced_service(
            name=name, namespace=self.namespace, body=body
        )
        logger.debug("Service `{}` was patched".format(name))
        return resp

    def create_or_update_service(self, name, body, reraise=False):
        try:
            return self.create_service(name=name, body=body), True
        except ApiException:
            try:
                return self.update_service(name=name, body=body), False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_pod(self, name, body):
        resp = self.k8s_api.create_namespaced_pod(namespace=self.namespace, body=body)
        logger.debug("Pod `{}` was created".format(name))
        return resp

    def update_pod(self, name, body):
        resp = self.k8s_api.patch_namespaced_pod(
            name=name, namespace=self.namespace, body=body
        )
        logger.debug("Pod `{}` was patched".format(name))
        return resp

    def create_or_update_pod(self, name, body, reraise=False):
        try:
            return self.create_pod(name=name, body=body), True
        except ApiException:
            try:
                return self.update_pod(name=name, body=body), False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_job(self, name, body):
        resp = self.k8s_batch_api.create_namespaced_job(
            namespace=self.namespace, body=body
        )
        logger.debug("Job `{}` was created".format(name))
        return resp

    def update_job(self, name, body):
        resp = self.k8s_batch_api.patch_namespaced_job(
            name=name, namespace=self.namespace, body=body
        )
        logger.debug("Job `{}` was patched".format(name))
        return resp

    def create_or_update_job(self, name, body, reraise=False):
        try:
            return self.create_job(name=name, body=body), True
        except ApiException:
            try:
                return self.update_job(name=name, body=body), False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_custom_object(self, name, group, version, plural, body):
        resp = self.k8s_custom_object_api.create_namespaced_custom_object(
            group=group,
            version=version,
            plural=plural,
            namespace=self.namespace,
            body=body,
        )
        logger.debug("Custom object `{}` was created".format(name))
        return resp

    def update_custom_object(self, name, group, version, plural, body):
        resp = self.k8s_custom_object_api.patch_namespaced_custom_object(
            name=name,
            group=group,
            version=version,
            plural=plural,
            namespace=self.namespace,
            body=body,
        )
        logger.debug("Custom object `{}` was patched".format(name))
        return resp

    def create_or_update_custom_object(
        self, name, group, version, plural, body, reraise=False
    ):
        try:
            return (
                self.create_custom_object(
                    name=name, group=group, version=version, plural=plural, body=body
                ),
                True,
            )

        except ApiException as e_create:
            try:
                return (
                    self.update_custom_object(
                        name=name,
                        group=group,
                        version=version,
                        plural=plural,
                        body=body,
                    ),
                    False,
                )
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError(
                        "Connection error: creation %s - update %s" % (e_create, e)
                    ) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_deployment(self, name, body):
        resp = self.k8s_apps_api.create_namespaced_deployment(
            namespace=self.namespace, body=body
        )
        logger.debug("Deployment `{}` was created".format(name))
        return resp

    def update_deployment(self, name, body):
        resp = self.k8s_apps_api.patch_namespaced_deployment(
            name=name, namespace=self.namespace, body=body
        )
        logger.debug("Deployment `{}` was patched".format(name))
        return resp

    def create_or_update_deployment(self, name, body, reraise=False):
        try:
            return self.create_deployment(name=name, body=body), True
        except ApiException:
            try:
                return self.update_deployment(name=name, body=body), False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_volume(self, name, body):
        resp = self.k8s_api.create_persistent_volume(body=body)
        logger.debug("Persistent volume `{}` was created".format(name))
        return resp

    def update_volume(self, name, body):
        resp = self.k8s_api.patch_persistent_volume(name=name, body=body)
        logger.debug("Persistent volume `{}` was patched".format(name))
        return resp

    def create_or_update_volume(self, name, body, reraise=False):
        try:
            return self.create_volume(name=name, body=body), True
        except ApiException:
            try:
                return self.update_service(name=name, body=body), False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_volume_claim(self, name, body):
        resp = self.k8s_api.create_namespaced_persistent_volume_claim(
            namespace=self.namespace, body=body
        )
        logger.debug("Volume claim `{}` was created".format(name))
        return resp

    def update_volume_claim(self, name, body):
        resp = self.k8s_api.patch_namespaced_persistent_volume_claim(
            name=name, namespace=self.namespace, body=body
        )
        logger.debug("Volume claim `{}` was patched".format(name))
        return resp

    def create_or_update_volume_claim(self, name, body, reraise=False):
        try:
            return self.create_volume_claim(name=name, body=body), True
        except ApiException:
            try:
                return self.update_volume_claim(name=name, body=body), False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def create_ingress(self, name, body):
        resp = self.networking_v1_beta1_api.create_namespaced_ingress(
            namespace=self.namespace, body=body
        )
        logger.debug("ingress `{}` was created".format(name))
        return resp

    def update_ingress(self, name, body):
        resp = self.networking_v1_beta1_api.patch_namespaced_ingress(
            name=name, namespace=self.namespace, body=body
        )
        logger.debug("Ingress `{}` was patched".format(name))
        return resp

    def create_or_update_ingress(self, name, body, reraise=False):
        try:
            return self.create_ingress(name=name, body=body), True
        except ApiException:
            try:
                return self.update_ingress(name=name, body=body), False
            except ApiException as e:
                if reraise:
                    raise PolyaxonK8sError("Connection error: %s" % e) from e
                else:
                    logger.error("K8S error: {}".format(e))

    def get_config_map(self, name, reraise=False):
        try:
            return self.k8s_api.read_namespaced_config_map(
                name=name, namespace=self.namespace
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def get_secret(self, name, reraise=False):
        try:
            return self.k8s_api.read_namespaced_secret(
                name=name, namespace=self.namespace
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def get_service(self, name, reraise=False):
        try:
            return self.k8s_api.read_namespaced_service(
                name=name, namespace=self.namespace
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def get_pod(self, name, reraise=False):
        try:
            return self.k8s_api.read_namespaced_pod(name=name, namespace=self.namespace)
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def get_job(self, name, reraise=False):
        try:
            return self.k8s_batch_api.read_namespaced_job(
                name=name, namespace=self.namespace
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def get_custom_object(self, name, group, version, plural):
        return self.k8s_custom_object_api.get_namespaced_custom_object(
            name=name,
            group=group,
            version=version,
            plural=plural,
            namespace=self.namespace,
        )

    def get_deployment(self, name, reraise=False):
        try:
            return self.k8s_apps_api.read_namespaced_deployment(
                name=name, namespace=self.namespace
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def get_volume(self, name, reraise=False):
        try:
            return self.k8s_api.read_persistent_volume(name=name)
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def get_volume_claim(self, name, reraise=False):
        try:
            return self.k8s_api.read_namespaced_persistent_volume_claim(
                name=name, namespace=self.namespace
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def get_ingress(self, name, reraise=False):
        try:
            return self.networking_v1_beta1_api.read_namespaced_ingress(
                name=name, namespace=self.namespace
            )
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            return None

    def delete_config_map(self, name, reraise=False):
        try:
            self.k8s_api.delete_namespaced_config_map(
                name=name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(api_version=constants.K8S_API_VERSION_V1),
            )
            logger.debug("Config map `{}` Deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Config map `{}` was not found".format(name))

    def delete_secret(self, name, reraise=False):
        try:
            self.k8s_api.delete_namespaced_secret(
                name=name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(api_version=constants.K8S_API_VERSION_V1),
            )
            logger.debug("secret `{}` Deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("secret `{}` was not found".format(name))

    def delete_service(self, name, reraise=False):
        try:
            self.k8s_api.delete_namespaced_service(
                name=name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(api_version=constants.K8S_API_VERSION_V1),
            )
            logger.debug("Service `{}` deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Service `{}` was not found".format(name))

    def delete_pod(self, name, reraise=False):
        try:
            self.k8s_api.delete_namespaced_pod(
                name=name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(api_version=constants.K8S_API_VERSION_V1),
            )
            logger.debug("Pod `{}` deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Pod `{}` was not found".format(name))

    def delete_job(self, name, reraise=False):
        try:
            self.k8s_batch_api.delete_namespaced_job(
                name=name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(api_version=constants.K8S_API_VERSION_V1),
            )
            logger.debug("Pod `{}` deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Pod `{}` was not found".format(name))

    def delete_custom_object(self, name, group, version, plural):
        self.k8s_custom_object_api.delete_namespaced_custom_object(
            name=name,
            group=group,
            version=version,
            plural=plural,
            namespace=self.namespace,
            body=client.V1DeleteOptions(),
        )
        logger.debug("Custom object `{}` deleted".format(name))

    def delete_deployment(self, name, reraise=False):
        try:
            self.k8s_apps_api.delete_namespaced_deployment(
                name=name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(
                    api_version=constants.K8S_API_VERSION_APPS_V1,
                    propagation_policy="Foreground",
                ),
            )
            logger.debug("Deployment `{}` deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Deployment `{}` was not found".format(name))

    def delete_volume(self, name, reraise=False):
        try:
            self.k8s_api.delete_persistent_volume(
                name=name,
                body=client.V1DeleteOptions(api_version=constants.K8S_API_VERSION_V1),
            )
            logger.debug("Volume `{}` Deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Volume `{}` was not found".format(name))

    def delete_volume_claim(self, name, reraise=False):
        try:
            self.k8s_api.delete_namespaced_persistent_volume_claim(
                name=name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(api_version=constants.K8S_API_VERSION_V1),
            )
            logger.debug("Volume claim `{}` Deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Volume claim `{}` was not found".format(name))

    def delete_ingress(self, name, reraise=False):
        try:
            self.networking_v1_beta1_api.delete_namespaced_ingress(
                name=name,
                namespace=self.namespace,
                body=client.V1DeleteOptions(
                    api_version=constants.K8S_API_VERSION_NETWORKING_V1_BETA1,
                    propagation_policy="Foreground",
                ),
            )
            logger.debug("Ingress `{}` deleted".format(name))
        except ApiException as e:
            if reraise:
                raise PolyaxonK8sError("Connection error: %s" % e) from e
            else:
                logger.debug("Ingress `{}` was not found".format(name))

    def delete_pods(self, include_uninitialized=True, reraise=False, **kwargs):
        objs = self.list_pods(
            include_uninitialized=include_uninitialized, reraise=reraise, **kwargs
        )
        for obj in objs:
            self.delete_pod(name=obj.metadata.name, reraise=reraise)

    def delete_jobs(self, include_uninitialized=True, reraise=False, **kwargs):
        objs = self.list_jobs(
            include_uninitialized=include_uninitialized, reraise=reraise, **kwargs
        )
        for obj in objs:
            self.delete_job(name=obj.metadata.name, reraise=reraise)

    def delete_services(self, reraise=False, **kwargs):
        objs = self.list_services(reraise=reraise, **kwargs)
        for obj in objs:
            self.delete_service(name=obj.metadata.name, reraise=reraise)

    def delete_deployments(self, reraise=False, **kwargs):
        objs = self.list_deployments(reraise=reraise, **kwargs)
        for obj in objs:
            self.delete_deployment(name=obj.metadata.name, reraise=reraise)

    def delete_ingresses(self, reraise=False, **kwargs):
        objs = self.list_services(reraise=reraise, **kwargs)
        for obj in objs:
            self.delete_ingress(name=obj.metadata.name, reraise=reraise)

    def is_pod_running(self, pod_id: str, container_id: str):
        event = self.k8s_api.read_namespaced_pod_status(pod_id, self.namespace)
        return is_pod_running(event, container_id)



import datetime

from typing import Optional

import aiofiles

from clipped.utils.paths import check_or_create_path, delete_path
from kubernetes_asyncio.client.models import V1Pod

from polyaxon._contexts import paths as ctx_paths
from polyaxon._k8s.logging.async_monitor import query_k8s_pod_logs
from polyaxon._k8s.manager.async_manager import AsyncK8sManager
from traceml.logging import V1Logs


async def sync_logs(
    run_uuid: str,
    k8s_manager: AsyncK8sManager,
    pod: V1Pod,
    last_time: Optional[datetime.datetime],
    stream: bool = False,
    is_running: bool = True,
):
    path_from = ctx_paths.CONTEXT_MOUNT_ARTIFACTS_FORMAT.format(run_uuid)
    path_from = "{}/.tmpplxlogs".format(path_from)

    if not is_running:
        delete_path(path_from)
        return

    logs, _ = await query_k8s_pod_logs(
        k8s_manager=k8s_manager,
        pod=pod,
        last_time=last_time,
        stream=stream,
    )
    if not logs:
        return

    path_from = "{}/{}.plx".format(path_from, pod.metadata.name)
    check_or_create_path(path_from, is_dir=False)
    async with aiofiles.open(path_from, "w") as filepath:
        _logs = V1Logs.construct(logs=logs)
        await filepath.write("{}\n{}".format(_logs.get_csv_header(), _logs.to_csv()))



from typing import Optional

from clipped.compact.pydantic import Field, StrictStr
from clipped.types.ref_or_obj import IntOrRef

from polyaxon._schemas.base import BaseSchemaModel


class V1SchedulingPolicy(BaseSchemaModel):
    _IDENTIFIER = "schedulingPolicy"

    min_available: Optional[IntOrRef] = Field(alias="minAvailable")
    queue: Optional[StrictStr]
    priority_class: Optional[StrictStr] = Field(alias="priorityClass")



from clipped.types.gcs import GcsPath as _GcsPath
from clipped.types.s3 import S3Path as _S3Path
from clipped.types.uri import Uri as _Uri
from clipped.types.wasb import WasbPath as _WasbPath


class GcsPath(_GcsPath):
    """GCS type.

    Args:
        bucket: str
        blob: str

    ### YAML usage

    The inputs definition

    ```yaml
    >>> inputs:
    >>>   - name: test1
    >>>     type: gcs
    >>>   - name: test2
    >>>     type: gcs
    ```

    The params usage

    ```yaml
    >>> params:
    >>>   test1: {value: "gs://bucket1"}
    >>>   test1: {value: "gs://bucket2/blobName"}
    ```

    ### Python usage

    The inputs definition

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="test1",
    >>>         type=types.GCS,
    >>>     ),
    >>>     V1IO(
    >>>         name="test2",
    >>>         type=types.GCS,
    >>>     ),
    >>> ]
    ```

    The params usage

    ```python
    >>> from polyaxon.schemas import V1Param
    >>> params = {
    >>>     "test1": V1Param(value="gs://bucket1"),
    >>>     "test2": V1Param(value="gs://bucket1/blobName")),
    >>> }
    ```
    """


class S3Path(_S3Path):
    """S3 type.

    Args:
        bucket: str
        key: str

    ### YAML usage

    The inputs definition

    ```yaml
    >>> inputs:
    >>>   - name: test1
    >>>     type: s3
    >>>   - name: test2
    >>>     type: s3
    ```

    The params usage

    ```yaml
    >>> params:
    >>>   test1: {value: "s3://bucket1"}
    >>>   test1: {value: "s3://bucket2/keyName"}}
    ```

    ### Python usage

    The inputs definition

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="test1",
    >>>         type=types.S3,
    >>>     ),
    >>>     V1IO(
    >>>         name="test2",
    >>>         type=types.S3,
    >>>     ),
    >>> ]
    ```

    The params usage

    ```python
    >>> from polyaxon.schemas import V1Param
    >>> params = {
    >>>     "test1": V1Param(value="s3://bucket1"),
    >>>     "test2": V1Param(value="s3://bucket1/keyName"),
    >>> }
    ```
    """


class WasbPath(_WasbPath):
    """Wasb type.

    Args:
        container: str
        storage_account: str
        path: str

    ### YAML usage

    The inputs definition

    ```yaml
    >>> inputs:
    >>>   - name: test1
    >>>     type: wasb
    >>>   - name: test2
    >>>     type: wasb
    ```

    The params usage

    ```yaml
    >>> params:
    >>>   test1: {value: "wasbs://container@user.blob.core.windows.net/path"}
    >>>   test2: {value: "wasbs://container@user.blob.core.windows.net/path"}
    ```

    ### Python usage

    The inputs definition

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="test1",
    >>>         type=types.WASB,
    >>>     ),
    >>>     V1IO(
    >>>         name="test2",
    >>>         type=types.WASB,
    >>>     ),
    >>> ]
    ```

    The params usage

    ```python
    >>> from polyaxon.schemas import V1Param
    >>> params = {
    >>>     "test1": V1Param(value="wasbs://container@user.blob.core.windows.net/path"),
    >>>     "test2": V1Param(value="wasbs://container@user.blob.core.windows.net/path"),
    >>> }
    ```
    """


class Uri(_Uri):
    """Uri type.

    Args:
        user: str
        password: str
        host: str

    ### YAML usage

    The inputs definition

    ```yaml
    >>> inputs:
    >>>   - name: test1
    >>>     type: uri
    >>>   - name: test2
    >>>     type: uri
    ```

    The params usage

    ```yaml
    >>> params:
    >>>   test1: {value: "https://username1:password1@service.com"}
    >>>   test1: {value: "https://username2:password2@service.com"}
    ```

    ### Python usage

    The inputs definition

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="test1",
    >>>         type=types.URI,
    >>>     ),
    >>> ]
    ```

    The params usage

    ```python
    >>> from polyaxon.schemas import V1Param
    >>> params = {
    >>>     "test1": V1Param(value="https://username2:password2@service.com"),
    >>> }
    ```

    > Normally you should not be passing auth details in plain values.
    """


# Backwards compatibility
V1GcsType = GcsPath
V1S3Type = S3Path
V1UriType = Uri
V1WasbType = WasbPath



from typing import Optional
from typing_extensions import Annotated

from clipped.compact.pydantic import Field, StrictStr, validate_arguments

from polyaxon._sdk.base_api import BaseApi
from polyaxon.exceptions import ApiTypeError


class ArtifactsStoresV1Api(BaseApi):
    @validate_arguments
    def upload_artifact(
        self,
        owner: Annotated[StrictStr, Field(..., description="Owner of the namespace")],
        uuid: Annotated[
            StrictStr, Field(..., description="Unique integer identifier of the entity")
        ],
        uploadfile: Annotated[StrictStr, Field(..., description="The file to upload.")],
        path: Annotated[
            Optional[StrictStr], Field(description="File path query params.")
        ] = None,
        overwrite: Annotated[
            Optional[bool], Field(description="File path query params.")
        ] = None,
        **kwargs
    ) -> None:
        """Upload artifact to a store

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.upload_artifact(owner, uuid, uploadfile, path, overwrite, async_req=True)
        >>> result = thread.get()

        :param owner: Owner of the namespace (required)
        :type owner: str
        :param uuid: Unique integer identifier of the entity (required)
        :type uuid: str
        :param uploadfile: The file to upload. (required)
        :type uploadfile: str
        :param path: File path query params.
        :type path: str
        :param overwrite: File path query params.
        :type overwrite: bool
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: None
        """
        kwargs["_return_http_data_only"] = True
        return self.upload_artifact_with_http_info(
            owner, uuid, uploadfile, path, overwrite, **kwargs
        )

    @validate_arguments
    def upload_artifact_with_http_info(
        self,
        owner: Annotated[StrictStr, Field(..., description="Owner of the namespace")],
        uuid: Annotated[
            StrictStr, Field(..., description="Unique integer identifier of the entity")
        ],
        uploadfile: Annotated[StrictStr, Field(..., description="The file to upload.")],
        path: Annotated[
            Optional[StrictStr], Field(description="File path query params.")
        ] = None,
        overwrite: Annotated[
            Optional[bool], Field(description="File path query params.")
        ] = None,
        **kwargs
    ):
        """Upload artifact to a store

        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True

        >>> thread = api.upload_artifact_with_http_info(owner, uuid, uploadfile, path, overwrite, async_req=True)
        >>> result = thread.get()

        :param owner: Owner of the namespace (required)
        :type owner: str
        :param uuid: Unique integer identifier of the entity (required)
        :type uuid: str
        :param uploadfile: The file to upload. (required)
        :type uploadfile: str
        :param path: File path query params.
        :type path: str
        :param overwrite: File path query params.
        :type overwrite: bool
        :param async_req: Whether to execute the request asynchronously.
        :type async_req: bool, optional
        :param _return_http_data_only: response data without head status code
                                       and headers
        :type _return_http_data_only: bool, optional
        :param _preload_content: if False, the urllib3.HTTPResponse object will
                                 be returned without reading/decoding response
                                 data. Default is True.
        :type _preload_content: bool, optional
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the authentication
                              in the spec for a single request.
        :type _request_auth: dict, optional
        :type _content_type: string, optional: force content-type for the request
        :return: Returns the result object.
                 If the method is called asynchronously,
                 returns the request thread.
        :rtype: None
        """

        _params = locals()

        _all_params = ["owner", "uuid", "uploadfile", "path", "overwrite"]
        _all_params.extend(
            [
                "async_req",
                "_return_http_data_only",
                "_preload_content",
                "_request_timeout",
                "_request_auth",
                "_content_type",
                "_headers",
            ]
        )

        # validate the arguments
        for _key, _val in _params["kwargs"].items():
            if _key not in _all_params:
                raise ApiTypeError(
                    "Got an unexpected keyword argument '%s'"
                    " to method upload_artifact" % _key
                )
            _params[_key] = _val
        del _params["kwargs"]

        _collection_formats = {}

        # process the path parameters
        _path_params = {}
        if _params["owner"]:
            _path_params["owner"] = _params["owner"]
        if _params["uuid"]:
            _path_params["uuid"] = _params["uuid"]

        # process the query parameters
        _query_params = []
        if _params.get("path") is not None:
            _query_params.append(("path", _params["path"]))
        if _params.get("overwrite") is not None:
            _query_params.append(("overwrite", _params["overwrite"]))

        # process the header parameters
        _header_params = dict(_params.get("_headers", {}))

        # process the form parameters
        _form_params = []
        _files = {}
        if _params["uploadfile"]:
            _files["uploadfile"] = _params["uploadfile"]

        # process the body parameter
        _body_params = None

        # set the HTTP header `Accept`
        _header_params["Accept"] = self.api_client.select_header_accept(
            ["application/json"]
        )

        # set the HTTP header `Content-Type`
        _content_types_list = _params.get(
            "_content_type",
            self.api_client.select_header_content_type(["multipart/form-data"]),
        )
        if _content_types_list:
            _header_params["Content-Type"] = _content_types_list

        # authentication setting
        _auth_settings = ["ApiKey"]

        _response_types_map = {}

        return self.api_client.call_api(
            "/api/v1/catalogs/{owner}/artifacts/{uuid}/upload",
            "POST",
            _path_params,
            _query_params,
            _header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            response_types_map=_response_types_map,
            auth_settings=_auth_settings,
            async_req=_params.get("async_req"),
            _return_http_data_only=_params.get("_return_http_data_only"),
            _preload_content=_params.get("_preload_content", True),
            _request_timeout=_params.get("_request_timeout"),
            collection_formats=_collection_formats,
            _request_auth=_params.get("_request_auth"),
        )



from typing import List, Optional, Union

from clipped.compact.pydantic import StrictStr
from clipped.types.ref_or_obj import BoolOrRef, IntOrRef, RefField

from polyaxon._flow.cache.enums import CacheSection
from polyaxon._schemas.base import BaseSchemaModel


class V1Cache(BaseSchemaModel):
    """Polyaxon provides a caching layer for operation executions,
    this behavior is enabled by default for all runs executed in the context of a DAG,
    a hyperparameter tuning, or a mapping.

    When runs are cached their outputs will be reused for future
    runs with similar inputs and component version.

    Args:
        disable: bool, optional, default: False
        ttl: int, optional
        io: List[str], optional
        sections: List[str], optional

    ## YAML usage

    ```yaml
    >>> cache:
    >>>   disable:
    >>>   ttl:
    >>>   inputs:
    >>>   sections:
    ```

    ## Python usage

    ```python
    >>> from polyaxon.schemas import V1Cache
    >>> cache = V1Cache(
    >>>   disable=False,
    >>>   ttl=3600,
    >>>   io=['param1', 'param4']
    >>>   sections=['init']
    >>> )
    ```

    ## Fields

    ### disable

    Caching is enabled by default, if you want to disable the cache
    for a component or just for a specific component run, you can set this field to `false`

    ```yaml
    >>> cache:
    >>>   disable: true
    ```

    ### ttl

    the default caching behavior is to persist and reuse a run's results everytime a new operation
    with similar characteristics is scheduled to run.

    In order to invalidate the cache after a certain period of time you can
    define a time to live value.

    ```yaml
    >>> cache:
    >>>   ttl: 36000  # 10 hours
    ```

    ### io

    You may want to discard an input/output from being considered for
    the cache state calculation,
    or you may want to cache a component's run irrespective of the params you pass to some io.

    This field gives you full control to define how you want to calculate the cache state.

    ```yaml
    >>> cache:
    >>>   io: ['param1', 'param4']
    ```

    ### sections

    By default the cache manager will consider the state of the `init`, `connections`,
    and `containers` (command and args) to trigger the cache hit logic.
    You may want to discard a section from being considered for
    the cache state calculation.

    This field gives you allows to define the sections that should be used
    to calculate the cache state.

    ```yaml
    >>> cache:
    >>>   sections: ['containers']
    ```
    """

    _IDENTIFIER = "cache"

    disable: Optional[BoolOrRef]
    ttl: Optional[IntOrRef]
    io: Optional[Union[List[StrictStr], RefField]]
    sections: Optional[Union[List[CacheSection], RefField]]



from typing_extensions import Literal

from polyaxon._flow.run.enums import V1RunKind
from polyaxon._flow.run.job import V1Job


class V1TunerJob(V1Job):
    _IDENTIFIER = V1RunKind.TUNER

    kind: Literal[_IDENTIFIER] = _IDENTIFIER



from clipped.utils.enums import PEnum

from polyaxon._schemas.lifecycle import V1Statuses


class V1EventKind(str, PEnum):
    RUN_STATUS_CREATED = "run_status_created"
    RUN_STATUS_RESUMING = "run_status_resuming"
    RUN_STATUS_COMPILED = "run_status_compiled"
    RUN_STATUS_ON_SCHEDULE = "run_status_on_schedule"
    RUN_STATUS_QUEUED = "run_status_queued"
    RUN_STATUS_SCHEDULED = "run_status_scheduled"
    RUN_STATUS_STARTING = "run_status_starting"
    RUN_STATUS_RUNNING = "run_status_running"
    RUN_STATUS_PROCESSING = "run_status_processing"
    RUN_STATUS_STOPPING = "run_status_stopping"
    RUN_STATUS_FAILED = "run_status_failed"
    RUN_STATUS_STOPPED = "run_status_stopped"
    RUN_STATUS_SUCCEEDED = "run_status_succeeded"
    RUN_STATUS_SKIPPED = "run_status_skipped"
    RUN_STATUS_WARNING = "run_status_warning"
    RUN_STATUS_UNSCHEDULABLE = "run_status_unschedulable"
    RUN_STATUS_UPSTREAM_FAILED = "run_status_upstream_failed"
    RUN_STATUS_RETRYING = "run_status_retrying"
    RUN_STATUS_UNKNOWN = "run_status_unknown"
    RUN_STATUS_DONE = "run_status_done"
    RUN_APPROVED_ACTOR = "run_approved_actor"
    RUN_INVALIDATED_ACTOR = "run_invalidated_actor"
    RUN_NEW_ARTIFACTS = "run_new_artifacts"
    CONNECTION_GIT_COMMIT = "connection_git_commit"
    CONNECTION_DATASET_VERSION = "connection_dataset_version"
    CONNECTION_REGISTRY_IMAGE = "connection_registry_image"
    ALERT_INFO = "alert_info"
    ALERT_WARNING = "alert_warning"
    ALERT_CRITICAL = "alert_critical"
    MODEL_VERSION_NEW_METRIC = "model_version_new_metric"
    PROJECT_CUSTOM_EVENT = "project_custom_event"
    ORG_CUSTOM_EVENT = "org_custom_event"

    @classmethod
    def get_events_statuses_mapping(cls):
        return {
            cls.RUN_STATUS_CREATED: V1Statuses.CREATED,
            cls.RUN_STATUS_RESUMING: V1Statuses.RESUMING,
            cls.RUN_STATUS_ON_SCHEDULE: V1Statuses.ON_SCHEDULE,
            cls.RUN_STATUS_COMPILED: V1Statuses.COMPILED,
            cls.RUN_STATUS_QUEUED: V1Statuses.QUEUED,
            cls.RUN_STATUS_SCHEDULED: V1Statuses.SCHEDULED,
            cls.RUN_STATUS_STARTING: V1Statuses.STARTING,
            cls.RUN_STATUS_RUNNING: V1Statuses.RUNNING,
            cls.RUN_STATUS_PROCESSING: V1Statuses.PROCESSING,
            cls.RUN_STATUS_STOPPING: V1Statuses.STOPPING,
            cls.RUN_STATUS_FAILED: V1Statuses.FAILED,
            cls.RUN_STATUS_STOPPED: V1Statuses.STOPPED,
            cls.RUN_STATUS_SUCCEEDED: V1Statuses.SUCCEEDED,
            cls.RUN_STATUS_SKIPPED: V1Statuses.SKIPPED,
            cls.RUN_STATUS_WARNING: V1Statuses.WARNING,
            cls.RUN_STATUS_UNSCHEDULABLE: V1Statuses.UNSCHEDULABLE,
            cls.RUN_STATUS_UPSTREAM_FAILED: V1Statuses.UPSTREAM_FAILED,
            cls.RUN_STATUS_RETRYING: V1Statuses.RETRYING,
            cls.RUN_STATUS_UNKNOWN: V1Statuses.UNKNOWN,
            cls.RUN_STATUS_DONE: V1Statuses.DONE,
        }



import re

from typing import Any, Dict, List, Optional, Union

from clipped.compact.pydantic import Field, StrictStr, root_validator, validator
from clipped.config.schema import skip_partial
from clipped.types.numbers import StrictIntOrFloat
from clipped.types.ref_or_obj import BoolOrRef, IntOrRef, RefField

from polyaxon._config.parser import ConfigParser
from polyaxon._schemas.base import BaseSchemaModel
from polyaxon.exceptions import PolyaxonSchemaError, PolyaxonValidationError

IO_NAME_BLACK_LIST = ["globals", "params", "connections"]
IO_NAME_ERROR = (
    "Received an input/output with a name in {}, "
    "please use a different name that does "
    "not already taken by the context".format(IO_NAME_BLACK_LIST)
)


def validate_io_value(
    name: str,
    type: str,
    value: Any,
    default: Any,
    is_optional: bool,
    is_list: bool,
    validation: Optional[Union["V1Validation", Dict]],
    parse: bool = True,
):
    try:
        parsed_value = ConfigParser.parse(type)(
            key=name,
            value=value,
            is_list=is_list,
            is_optional=is_optional,
            default=default,
        )
        if validation:
            validation.run_validation(
                value=parsed_value, type=type, is_optional=is_optional
            )
        if parse:
            return parsed_value
        # Return the original value, the parser will return specs sometimes
        if value is not None:
            return value
        return default
    except PolyaxonSchemaError as e:
        raise PolyaxonValidationError(
            "Could not parse value `%s` for `%s`, an error was encountered: %s"
            % (value, name, e)
        )


def validate_io(
    name: str,
    type: str,
    value: Any,
    is_optional: bool,
    is_list: bool,
    is_flag: bool,
    validation: Optional[Union["V1Validation", Dict]],
):
    if type and value:
        try:
            value = validate_io_value(
                name=name,
                type=type,
                value=value,
                default=None,
                is_list=is_list,
                is_optional=is_optional,
                validation=validation,
            )
        except PolyaxonValidationError as e:
            raise ValueError(e)

    if not is_optional and value:
        raise ValueError(
            "IO `{}` is not optional and has default value `{}`. "
            "Please either make it optional or remove the default value.".format(
                name, value
            )
        )

    if is_flag and type != "bool":
        raise TypeError(
            "IO type `{}` cannot be a flag, it must be of type `{}`".format(
                type, "bool"
            )
        )

    return value


class V1Validation(BaseSchemaModel):
    """Validation is used to validate inputs/outputs.

    Validation is defined as a sdt of predicates, each one of the predicates that must be satisfied.

    Args:
        delay: bool, optional
        gt: int, optional
        ge: int, optional
        lt: int, optional
        le: int, optional
        multiple_of: int, optional
        min_items: int, optional
        max_digits: int, optional
        decimal_places: int, optional
        regex: str, optional
        min_length: int, optional
        max_length: int, optional
        keys: List[str], optional
        contains_keys: List[str], optional
        excludes_keys: List[str], optional
        contains: List[str], optional
        excludes: List[str], optional
        options: List[str], optional

    ## YAML usage

    ```yaml
    >>> inputs:
    >>>   - name: loss
    >>>     type: str
    >>>     validation:
    >>>       options: [MeanSquaredError, MeanAbsoluteError]
    >>>   - name: learning_rate
    >>>     type: float
    >>>     validation:
    >>>       gt: 0.001
    >>>       lt: 0.5
    >>> options:
    >>>   - name: accuracy
    >>>     type: float
    >>>     validation:
    >>>       ge: 0.5
    >>>   - name: outputs-path
    >>>     type: path
    >>>     validation:
    >>>       regex: "^s3://(?P<bucket>[a-z0-9-.]{3,63})/(?P<key>.+)$"
    ```

    ## Python usage

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import V1Validation
    >>> inputs = [
    >>>     V1IO(
    >>>         name="loss",
    >>>         type='str',
    >>>         validation=[V1Validation(options=["MeanSquaredError", "MeanAbsoluteError"])]
    >>>     ),
    >>>     V1IO(
    >>>         name="learning_rate",
    >>>         type='float',
    >>>         validation=[V1Validation(gt=0.001, lt=0.5)]
    >>> ]
    >>> outputs = [
    >>>     V1IO(
    >>>         name="accuracy",
    >>>         type='float',
    >>>         validation=[V1Validation(ge=0.5)]
    >>>     ),
    >>>     V1IO(
    >>>         name="outputs-path",
    >>>         type=types.PATH,
    >>>         validation=[V1Validation(regex="^s3://(?P<bucket>[a-z0-9-.]{3,63})/(?P<key>.+)$")]
    >>>     )
    >>> ]
    ```

    ## Validation

    ### Delay

    To instruct the parser to only validate the input/output at compilation or resolution time:

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    >>>     validation:
    >>>       delay: true
    ```

    This flag is enabled by default for outputs, since they can only be
    resolved after or during the run. To request validation at compilation time for outputs,
    you need to set this flag to `False`.

    ### Validators

    #### Numerical Constraints

    > these constraints are also applied item wise for lists and for dict values

    * gt - greater than
    * lt - less than
    * ge - greater than or equal to
    * le - less than or equal to
    * multipleOf - a multiple of the given number

    #### Decimal Constraints

    > these constraints are also applied item wise for lists and for dict values

    * minDigits - minimum number of digits
    * maxDigits - maximum number of digits
    * decimalPlaces - maximum number of decimal places

    #### String Constraints

    > these constraints are also applied item wise for lists and for dict values

    * regex - a regex pattern
    * minLength - minimum length
    * maxLength - maximum length

    #### Generic Constraints

    > these constraints are also applied item wise for lists and for dict values

    * contains - a list of values that must be present
    * excludes - a list of values that must not be present
    * options - a list of values that must be present

    #### Dict Constraints
    * keys - a list of keys that must be present in the dict
    * containsKeys - a list of keys that must be present in the dict
    * excludesKeys - a list of keys that must not be present in the dict

    #### Items Constraints

    * minItems - minimum number of items
    * maxItems - maximum number of items
    """

    delay: Optional[BoolOrRef]
    gt: Optional[Union[StrictIntOrFloat, RefField]]
    ge: Optional[Union[StrictIntOrFloat, RefField]]
    lt: Optional[Union[StrictIntOrFloat, RefField]]
    le: Optional[Union[StrictIntOrFloat, RefField]]
    multiple_of: Optional[Union[StrictIntOrFloat, RefField]] = Field(alias="multipleOf")
    min_digits: Optional[IntOrRef] = Field(alias="minDigits")
    max_digits: Optional[IntOrRef] = Field(alias="maxDigits")
    decimal_places: Optional[IntOrRef] = Field(alias="decimalPlaces")
    regex: Optional[StrictStr]
    min_length: Optional[IntOrRef] = Field(alias="minLength")
    max_length: Optional[IntOrRef] = Field(alias="maxLength")
    contains: Optional[Any]
    excludes: Optional[Any]
    options: Optional[Any]
    min_items: Optional[IntOrRef] = Field(alias="minItems")
    max_items: Optional[IntOrRef] = Field(alias="maxItems")
    keys: Optional[Union[List[StrictStr], RefField]]
    contains_keys: Optional[Union[List[StrictStr], RefField]] = Field(
        alias="containsKeys"
    )
    excludes_keys: Optional[Union[List[StrictStr], RefField]] = Field(
        alias="excludesKeys"
    )

    def _validate_gt(self, value):
        if self.gt is not None and value <= self.gt:
            raise PolyaxonValidationError(
                f"Value `{value}` must be greater than `{self.gt}`"
            )

    def _validate_ge(self, value):
        if self.ge is not None and value < self.ge:
            raise PolyaxonValidationError(
                f"Value `{value}` must be greater than or equal to `{self.ge}`"
            )

    def _validate_lt(self, value):
        if self.lt is not None and value >= self.lt:
            raise PolyaxonValidationError(
                f"Value `{value}` must be less than `{self.lt}`"
            )

    def _validate_le(self, value):
        if self.le is not None and value > self.le:
            raise PolyaxonValidationError(
                f"Value `{value}` must be less than or equal to `{self.le}`"
            )

    def _validate_multiple_of(self, value):
        if self.multiple_of is not None and value % self.multiple_of != 0:
            raise PolyaxonValidationError(
                f"Value `{value}` must be a multiple of `{self.multiple_of}`"
            )

    def _validate_min_digits(self, value):
        if (
            self.min_digits is not None
            and len(str(value).replace(".", "")) < self.min_digits
        ):
            raise PolyaxonValidationError(
                f"Value `{value}` must have at least `{self.min_digits}` digits"
            )

    def _validate_max_digits(self, value):
        if (
            self.max_digits is not None
            and len(str(value).replace(".", "")) > self.max_digits
        ):
            raise PolyaxonValidationError(
                f"Value `{value}` must have at most `{self.max_digits}` digits"
            )

    def _validate_decimal_places(self, value):
        if (
            self.decimal_places is not None
            and len(str(value).split(".")[1]) > self.decimal_places
        ):
            raise PolyaxonValidationError(
                f"Value `{value}` must have at most `{self.decimal_places}` decimal places"
            )

    def _validate_regex(self, value):
        if self.regex is not None and not re.match(self.regex, value):
            raise PolyaxonValidationError(
                f"Value `{value}` must match the regex `{self.regex}`"
            )

    def _validate_min_length(self, value):
        if self.min_length is not None and len(value) < self.min_length:
            raise PolyaxonValidationError(
                f"Value `{value}` must have at least `{self.min_length}` characters"
            )

    def _validate_max_length(self, value):
        if self.max_length is not None and len(value) > self.max_length:
            raise PolyaxonValidationError(
                f"Value `{value}` must have at most `{self.max_length}` characters"
            )

    def _validate_contains(self, value):
        if self.contains is not None and self.contains not in value:
            raise PolyaxonValidationError(
                f"Value `{value}` must contain one of the values `{self.contains}`"
            )

    def _validate_excludes(self, value):
        if self.excludes is not None and self.excludes in value:
            raise PolyaxonValidationError(
                f"Value `{value}` must not contain any of the values `{self.excludes}`"
            )

    def _validate_options(self, value):
        if self.options is not None and value not in self.options:
            raise PolyaxonValidationError(
                f"Value `{value}` must be one of the values `{self.options}`"
            )

    def _validate_min_items(self, value):
        if self.min_items is not None and len(value) < self.min_items:
            raise PolyaxonValidationError(
                f"Value `{value}` must have at least `{self.min_items}` items"
            )

    def _validate_max_items(self, value):
        if self.max_items is not None and len(value) > self.max_items:
            raise PolyaxonValidationError(
                f"Value `{value}` must have at most `{self.max_items}` items"
            )

    def _validate_keys(self, value):
        if self.keys is not None and set(value.keys()) != set(self.keys):
            raise PolyaxonValidationError(
                f"Value `{value}` must contain all the keys `{self.keys}`"
            )

    def _validate_contains_keys(self, value):
        if self.contains_keys is not None and not all(
            k in value for k in self.contains_keys
        ):
            raise PolyaxonValidationError(
                f"Value `{value}` must contain all the keys `{self.contains_keys}`"
            )

    def _validate_excludes_keys(self, value):
        if self.excludes_keys is not None and any(
            k in value for k in self.excludes_keys
        ):
            raise PolyaxonValidationError(
                f"Value `{value}` must not contain any of the keys `{self.excludes_keys}`"
            )

    def run_validation(
        self,
        value: Any,
        type: str,
        is_optional: bool,
    ):
        if value is None and is_optional:
            return

        def _validate_value(v):
            self._validate_gt(v)
            self._validate_ge(v)
            self._validate_lt(v)
            self._validate_le(v)
            self._validate_multiple_of(v)
            self._validate_min_digits(v)
            self._validate_max_digits(v)
            self._validate_decimal_places(v)
            self._validate_regex(v)
            self._validate_min_length(v)
            self._validate_max_length(v)
            self._validate_contains(v)
            self._validate_excludes(v)
            self._validate_options(v)

        if isinstance(value, (list, tuple, set)):
            for v in value:
                _validate_value(v)
        elif isinstance(value, dict):
            for v in value.values():
                _validate_value(v)
        else:
            _validate_value(value)

        if isinstance(value, (list, tuple, set, dict)):
            self._validate_min_items(value)
            self._validate_max_items(value)

        if isinstance(value, dict):
            self._validate_keys(value)
            self._validate_contains_keys(value)
            self._validate_excludes_keys(value)


class V1IO(BaseSchemaModel):
    """Each Component may have its own inputs and outputs.
    The inputs and outputs describe the expected parameters to pass to the component
    and their types. In the context of a DAG,
    inputs and outputs types are used to validate the flow of information
    going from one operation to another.

    The final value of an input/output can be resolved
    from [params](/docs/core/specification/params/), or from other values in
    the [context](/docs/core/context/).

    Examples:
     * A build component may have a git repository as input and a container image as output.
     * A training component may have a container image, data path,
       and some hyperparameters as input and a list of metrics and artifacts as outputs.

    An input/output section includes a name, a description, a type to check the value passed,
    a flag to tell if the input/output is optional, and a default value if it's optional.

    Sometimes users prefer to pass a param to the `command` or `args`
    section only if the value is not null. Polyaxon provides a way to do that using
    `{{ params.PARAM_NAME.as_arg }}`,
    this will be empty if the value is null or `--PARAM_NAME=value` if the value is not null.
    If the value is of type bool, it will be `--PARAM_NAME`.
    It's also possible to control how the param should be
    converted to an argument using the `arg_format`.

    To learn how to pass valid values,
    please check the [param section](/docs/core/specification/params/).

    Args:
        name: str
        description: str, optional
        type: str, any python type hint, pydantic built-in types, and gcs, s3, wasb, dockerfile, git, image, event, artifacts, path, metric, metadata, date, datetime.
        value: any, optional
        is_optional: bool, optional (**Deprecated**)
        is_list: bool, optional
        is_flag: bool, optional
        arg_format: str, optional
        connection: str, optional
        to_init: bool, optional
        to_env: str, optional
        validation: [V1Validation](/docs/core/specification/validation/), optional
        delay_validation: bool, optional (**Deprecated**: please see valiation.delay)
        options: List[any], optional (**Deprecated**: please see valiation)

    ## YAML usage

    ```yaml
    >>> inputs:
    >>>   - name: loss
    >>>     type: str
    >>>     isOptional: true
    >>>     value: MeanSquaredError
    >>>   - name: preprocess
    >>>     type: bool
    >>> outputs:
    >>>   - name: accuracy
    >>>     type: float
    >>>   - name: outputs-path
    >>>     type: path
    ```

    ## Python usage

    ```python
    >>> from polyaxon import types
    >>> from polyaxon.schemas import V1IO
    >>> inputs = [
    >>>     V1IO(
    >>>         name="loss",
    >>>         type='str',
    >>>         description="Loss to use for training my model",
    >>>         is_optional=True,
    >>>         value="MeanSquaredError"
    >>>     ),
    >>>     V1IO(
    >>>         name="preprocess",
    >>>         type='bool',
    >>>         description="A flag to preprocess data before training",
    >>>     )
    >>> ]
    >>> outputs = [
    >>>     V1IO(
    >>>         name="accuracy",
    >>>         type='float',
    >>>     ),
    >>>     V1IO(
    >>>         name="outputs-path",
    >>>         type=types.PATH,
    >>>     )
    >>> ]
    ```

    These inputs/outputs declarations can be used to pass values to our program:

    ```bash
     ... --loss={{ loss }} {{ params.preprocess.as_arg }}
    ```

    ## Fields

    ### name

    The input/output name. The name must be a valid slug, and cannot include dots `.`.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    ```

    ### description

    An optional description for the input/output.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    ```

    ### type

    The type of the input/output. The type will be used to validate the value

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    ```

    for more details about composite type validation and schema,
    please check the [types section](/docs/core/specification/types/),
    possible types include any python type hint, pydantic built-in types, and:
        * URI: "uri"
        * LIST: "list"
        * GCS: "gcs"
        * S3: "s3"
        * WASB: "wasb"
        * DOCKERFILE: "dockerfile"
        * GIT: "git"
        * IMAGE: "image"
        * EVENT: "event"
        * ARTIFACTS: "artifacts"
        * PATH: "path"
        * METRIC: "metric"
        * METADATA: "metadata"

    ### value

    If an input is optional you should assign it a value.
    If an output is optional you can assign it a value.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    ```

    ### isOptional

    A flag to tell if an input/output is optional.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    >>>     isOptional: true
    ```

    ### isList

    > **Deprecated**: Please use `type: List[TYPING]` instead.

    In `v2` you should:

    ```yaml
    >>> inputs:
    >>>   - name: learning_rates
    >>>     type: List[float]
    ```

    A flag used in `v1` to tell if an input/output is a list of the type passed.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rates
    >>>     type: float
    >>>     isList: true
    ```

    In this case the input name `learning_rates` will expect a value of type `List[float]`,
    e.g. [0.1 0.01, 0.0001]

    ### argFormat

    A key to control how to convert an input/output to a CLI argument if the value is not null,
    the default behavior:

     * For bool type: If the resolved value of the input is True,
       `"{{ params.PARAM_NAME.as_arg }}"` will be resolved to `"--PARAM_NAME"`
       otherwise it will be an empty string `""`.
     * For non-bool types: If the resolved value of the input is not null,
       `"{{ params.PARAM_NAME.as_arg }}"` will be resolved to `"--PARAM_NAME=value"`
       otherwise it will be an empty string `""`.

    Let's look at the flowing example:

    ```yaml
    >>> inputs:
    >>>   - name: lr
    >>>     type: float
    >>>   - name: check
    >>>     type: bool
    ```

    **This manifest**:

    ```yaml
    >>> container:
    >>>    command: ["{{ lr }}", "{{ check }}"]
    ```

    Will be transformed to:

    ```yaml
    >>> container:
    >>>    command: ["0.01", "true"]
    ```

    **This manifest**:

    ```yaml
    >>> container:
    >>>    command: ["{{ params.lr.as_arg }}", "{{ params.check.as_arg }}"]
    ```

    Will be transformed to:

    ```yaml
    >>> container:
    >>>    command: ["--lr=0.01", "--check"]
    ```

    Changing the behavior with `argFormat`:

    ```yaml
    >>> inputs:
    >>>   - name: lr
    >>>     type: float
    >>>     argFormat: "lr={{ lr }}"
    >>>   - name: check
    >>>     type: bool
    >>>     argFormat: "{{ 1 if check else 0 }}"
    ```

    **This manifest**:

    ```yaml
    >>> container:
    >>>    command: ["{{ params.lr.as_arg }}", "{{ params.check.as_arg }}"]
    ```

    Will be transformed to:

    ```yaml
    >>> container:
    >>>    command: ["lr=0.01", "1"]
    ```

    ### connection

    A connection to use with the input/outputs.

    ### toInit

    If True, it will be converted to an init container.

    ### toEnv

    > **N.B**: Requires Polyaxon CLI and Polyaxon Agent/CE version `>= 1.12`

    If passed, it will be converted automatically to an environment variable.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     type: float
    >>>     value: 1.1
    >>>     toEnv: MY_LEARNING_RATE
    ```

    ### validation [v2]

    A schema to use to validate the input/output value or to delay the validation to runtime.

    To instruct the parser to only validate the input/output at compilation or resolution time:

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    >>>     validation:
    >>>       delay: true
    ```

    This flag is enabled by default for outputs, since they can only be
    resolved after or during the run. To request validation at compilation time for outputs,
    you need to set this flag to `False`.

    To validate an input/output value:

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    >>>     validation:
    >>>       min: 0.01
    >>>       max: 0.1
    ```

    Options allow to pass a list of values that will be used to validate any passed params.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 0.1
    >>>     validation:
    >>>       options: [0.1, 0.2, 0.3]
    ```

    ### delayValidation

    > **Deprecated**: Please use `validation: ...` instead.

    A flag to tell if an input/output should not be
    validated at compilation or resolution time.

    This flag is enabled by default for outputs, since they can only be
    resolved after or during the run. To request validation at compilation time for outputs,
    you need to set this flag to `False`.


    ### options

    > **Deprecated**: Please use `validation: ...` instead.

    Options allow to pass a list of values that will be used to validate any passed params.

    ```yaml
    >>> inputs:
    >>>   - name: learning_rate
    >>>     description: A short description about this input
    >>>     type: float
    >>>     value: 1.1
    >>>     options: [1.1, 2.2, 3.3]
    ```

    If you pass the value `4.4` for the learning rate it will raise a validation error.

    ## Example


    ```yaml
    >>> version: 1.1
    >>> kind: component
    >>> inputs:
    >>>   - name: batch_size
    >>>     description: batch size
    >>>     isOptional: true
    >>>     value: 128
    >>>     type: int
    >>>   - name: num_steps
    >>>     isOptional: true
    >>>     default: 500
    >>>     type: int
    >>>   - name: learning_rate
    >>>     isOptional: true
    >>>     default: 0.001
    >>>     type: float
    >>>   - name: dropout
    >>>     isOptional: true
    >>>     default: 0.25
    >>>     type: float
    >>>   - name: num_epochs
    >>>     isOptional: true
    >>>     default: 1
    >>>     type: int
    >>>   - name: activation
    >>>     isOptional: true
    >>>     default: relu
    >>>     type: str
    >>> run:
    >>>   kind: job
    >>>   image: foo:bar
    >>>   container:
    >>>     command: [python3, model.py]
    >>>     args: [
    >>>         "--batch_size={{ batch_size }}",
    >>>         "--num_steps={{ num_steps }}",
    >>>         "--learning_rate={{ learning_rate }}",
    >>>         "--dropout={{ dropout }",
    >>>         "--num_epochs={{ num_epochs }}",
    >>>         "--activation={{ activation }}"
    >>>     ]
    ```

    ### Running a typed component using the CLI

    Using the Polyaxon CLI we can now run this component and override the inputs' default values:

    ```bash
    polyaxon run -f polyaxonfile.yaml -P activation=sigmoid -P dropout=0.4
    ```

    This will generate a manifest and will replace the params passed and validated against the inputs' types.

    ### Required inputs

    In the example all inputs are optional.
    If we decide for instance to make the activation required:

    ````yaml
    >>> ...
    >>> inputs:
    >>>   ...
    >>>   - name: activation
    >>>     type: str
    >>>   ...
    ...
    ````

    By changing this input, polyaxon can not run this component without passing the activation:


    ```bash
    polyaxon run -f polyaxonfile.yaml -P activation=sigmoid
    ```
    """

    _IDENTIFIER = "io"

    name: StrictStr
    description: Optional[StrictStr]
    type: Optional[StrictStr]
    is_optional: Optional[bool] = Field(alias="isOptional")
    is_list: Optional[bool] = Field(alias="isList")
    is_flag: Optional[bool] = Field(alias="isFlag")
    arg_format: Optional[StrictStr] = Field(alias="argFormat")
    connection: Optional[StrictStr]
    to_init: Optional[bool] = Field(alias="toInit")
    to_env: Optional[StrictStr] = Field(alias="toEnv")
    value: Optional[Any]
    validation: Optional[V1Validation]
    delay_validation: Optional[bool] = Field(alias="delayValidation")
    options: Optional[Any]

    @validator("name", always=True)
    def validate_name(cls, v):
        if v in IO_NAME_BLACK_LIST:
            raise ValueError(IO_NAME_ERROR)
        return v

    @root_validator(pre=True)
    def handle_validation(cls, values):
        validation = values.get("validation")
        if not validation and (
            values.get("options") is not None
            or values.get("delay_validation") is not None
        ):
            validation = {}
        if isinstance(validation, dict):
            validation = V1Validation(**validation)
        if values.get("options") is not None:
            validation.options = values.pop("options")
        if values.get("delay_validation") is not None:
            validation.delay = values.pop("delay_validation")
        if validation:
            values["validation"] = validation
        return values

    @root_validator
    @skip_partial
    def validate_io(cls, values):
        validate_io(
            name=values.get("name"),
            type=values.get("type"),
            value=values.get("value"),
            is_list=values.get("is_list"),
            is_optional=values.get("is_optional"),
            is_flag=values.get("is_flag"),
            validation=values.get("validation"),
        )
        return values

    def validate_value(self, value: Any, parse: bool = True):
        if self.type is None:
            return value

        return validate_io_value(
            name=self.name,
            type=self.type,
            value=value,
            default=self.value,
            is_list=self.is_list,
            is_optional=self.is_optional,
            validation=self.validation,
            parse=parse,
        )

    def get_repr_from_value(self, value):
        """A string representation that is used to create hash cache"""
        value = self.validate_value(value=value, parse=False)
        io_dict = self.to_light_dict(include_attrs=["name", "type"])
        io_dict["value"] = value
        return io_dict

    def get_repr(self):
        """A string representation that is used to create hash cache"""
        io_dict = self.to_light_dict(include_attrs=["name", "type", "value"])
        return io_dict



from clipped.utils.enums import PEnum


class V1RunKind(str, PEnum):
    JOB = "job"
    SERVICE = "service"
    DAG = "dag"
    DASKJOB = "daskjob"
    RAYJOB = "rayjob"
    MPIJOB = "mpijob"
    TFJOB = "tfjob"
    PYTORCHJOB = "pytorchjob"
    PADDLEJOB = "paddlejob"
    MXJOB = "mxjob"
    XGBJOB = "xgbjob"
    MATRIX = "matrix"
    SCHEDULE = "schedule"
    TUNER = "tuner"
    WATCHDOG = "watchdog"
    NOTIFIER = "notifier"
    CLEANER = "cleaner"
    BUILDER = "builder"

    @classmethod
    def has_service(cls, kind: str):
        return kind in (
            cls.SERVICE,
            cls.DASKJOB,
            cls.RAYJOB,
            cls.MPIJOB,
            cls.TFJOB,
            cls.PYTORCHJOB,
            cls.PADDLEJOB,
            cls.MXJOB,
            cls.XGBJOB,
        )


class V1CloningKind(str, PEnum):
    COPY = "copy"
    RESTART = "restart"
    CACHE = "cache"


class V1PipelineKind(str, PEnum):
    DAG = "dag"
    MATRIX = "matrix"


class V1RunEdgeKind(str, PEnum):
    ACTION = "action"
    EVENT = "event"
    HOOK = "hook"
    DAG = "dag"
    JOIN = "join"
    RUN = "run"
    TB = "tb"
    BUILD = "build"


class V1RunPending(str, PEnum):
    APPROVAL = "approval"
    UPLOAD = "upload"
    CACHE = "cache"
    BUILD = "build"



from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from clipped.compact.pydantic import Field, StrictStr, validator

from polyaxon._k8s import k8s_schemas, k8s_validation
from polyaxon._schemas.base import BaseSchemaModel

if TYPE_CHECKING:
    from polyaxon._flow import V1Environment


class V1DefaultScheduling(BaseSchemaModel):
    _IDENTIFIER = "default_scheduling"
    _SWAGGER_FIELDS = ["affinity", "tolerations"]

    node_selector: Optional[Dict[StrictStr, StrictStr]] = Field(alias="nodeSelector")
    affinity: Optional[Union[k8s_schemas.V1Affinity, Dict]]
    tolerations: Optional[List[Union[k8s_schemas.V1Toleration, Dict]]]
    image_pull_secrets: Optional[List[StrictStr]] = Field(alias="imagePullSecrets")

    @validator("affinity", always=True, pre=True)
    def validate_affinity(cls, v) -> k8s_schemas.V1Affinity:
        return k8s_validation.validate_k8s_affinity(v)

    @validator("tolerations", always=True, pre=True)
    def validate_tolerations(cls, v) -> List[k8s_schemas.V1Toleration]:
        if not v:
            return v
        return [k8s_validation.validate_k8s_toleration(vi) for vi in v]

    @staticmethod
    def get_service_environment(
        service: Any,
        default_scheduling: "V1DefaultScheduling",
    ) -> "V1Environment":
        from polyaxon._flow import V1Environment

        env = V1Environment.construct()
        if service and service.node_selector:
            env.node_selector = service.node_selector
        elif default_scheduling and default_scheduling.node_selector:
            env.node_selector = default_scheduling.node_selector
        if service and service.affinity:
            env.affinity = service.affinity
        elif default_scheduling and default_scheduling.affinity:
            env.affinity = default_scheduling.affinity
        if service and service.tolerations:
            env.tolerations = service.tolerations
        elif default_scheduling and default_scheduling.tolerations:
            env.tolerations = default_scheduling.tolerations
        if service and service.image_pull_secrets:
            env.image_pull_secrets = service.image_pull_secrets
        elif default_scheduling and default_scheduling.image_pull_secrets:
            env.image_pull_secrets = default_scheduling.image_pull_secrets

        return env



from typing import Dict, Optional

from polyaxon._k8s.converter.pod.spec import get_pod_spec, get_pod_template_spec
from polyaxon._k8s.replica import ReplicaSpec


def get_kf_replicas_template(
    namespace: str,
    resource_name: str,
    replica_name: str,
    replica: Optional[ReplicaSpec],
    labels: Dict[str, str],
    annotations: Dict[str, str],
    template_spec: Dict,
):
    if not replica:
        return

    metadata, pod_spec = get_pod_spec(
        namespace=namespace,
        main_container=replica.main_container,
        sidecar_containers=replica.sidecar_containers,
        init_containers=replica.init_containers,
        resource_name=resource_name,
        volumes=replica.volumes,
        environment=replica.environment,
        labels=labels,
        annotations=annotations,
    )

    template_spec[replica_name] = {
        "replicas": replica.num_replicas,
        "restartPolicy": pod_spec.restart_policy or "Never",
        "template": get_pod_template_spec(metadata=metadata, pod_spec=pod_spec),
    }



from typing import Union
from typing_extensions import Annotated

from clipped.compact.pydantic import Field

from polyaxon._flow.early_stopping.policies import (
    V1DiffStoppingPolicy,
    V1FailureEarlyStopping,
    V1MedianStoppingPolicy,
    V1MetricEarlyStopping,
    V1TruncationStoppingPolicy,
)

V1EarlyStopping = Annotated[
    Union[V1MetricEarlyStopping, V1FailureEarlyStopping],
    Field(discriminator="kind", alias="earlyStopping"),
]



from polyaxon._auxiliaries.cleaner import (
    V1PolyaxonCleaner,
    get_default_cleaner_container,
)
from polyaxon._auxiliaries.default_scheduling import V1DefaultScheduling
from polyaxon._auxiliaries.init import (
    V1PolyaxonInitContainer,
    get_default_init_container,
    get_init_resources,
)
from polyaxon._auxiliaries.notifier import (
    V1PolyaxonNotifier,
    get_default_notification_container,
)
from polyaxon._auxiliaries.sidecar import (
    V1PolyaxonSidecarContainer,
    get_default_sidecar_container,
    get_sidecar_resources,
)
from polyaxon._auxiliaries.tuner import get_default_tuner_container



from polyaxon._k8s.executor.base import BaseExecutor
from polyaxon._k8s.manager.manager import K8sManager


class Executor(BaseExecutor):
    def _get_manager(self):
        return K8sManager(
            k8s_config=self.k8s_config,
            namespace=self.namespace,
            in_cluster=self.in_cluster,
        )



from typing import Optional

from clipped.compact.pydantic import Field, StrictStr

from polyaxon._schemas.base import BaseSchemaModel


class UIConfig(BaseSchemaModel):
    enabled: Optional[bool]
    offline: Optional[bool]
    static_url: Optional[StrictStr] = Field(alias="staticUrl")
    base_url: Optional[StrictStr] = Field(alias="baseUrl")
    assets_version: Optional[StrictStr] = Field(alias="assetsVersion")
    admin_enabled: Optional[bool] = Field(alias="adminEnabled")
